import os
import json
import click
import time
import webbrowser
import pprint
from tempfile import NamedTemporaryFile
from max_ard.select import Select
import sys
from collections import defaultdict
from max_ard.exceptions import BadARDRequest, NotFinished

@click.group()
def select():
    ''' Tools for interacting with the ARD Select service '''
    pass

    

@click.command()
@click.argument('select_id', required=True)
@click.option('--format', help='Return the status in an alternate format. `raw` will return the full response JSON.')
def status(select_id, format):
    ''' Return the status of a Select request using it's SELECT_ID '''
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    if format == 'raw':
        print(s.response.dict())
    else:
        print(s.state)


@click.command()
@click.argument('select_id', required=True)
@click.option('--format', help='Returns the select in one of the generated format types: `html`, `geojson`, `geojsonl`, `stac`.')
def describe(select_id, format):
    ''' Returns info about the Select given the SELECT_ID. If no FORMAT is supplied, prints out a summary of the select.
        
        Providing a FORMAT returns the contents of the given Select result file generated by the system. 
        Formats `stac`, `geojson`, and `geojsonl` return files describing the tiles. Format `html` is an interactive map 
        browser of the results - to view the map without downloading it first, use `maxard select browse SELECT_ID` '''

    try:
        s = Select.from_id(select_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    if not s.finished:
        print('Select has not finished running')
        print('*******************')
        print('Request details:')
        print(json.dumps(s.request.to_payload()))
        print('*******************')
        exit()
    if s.response.error_message is not None:
        error = s.response.error_message
        print(f'Error in selection process: {error["Error"]}')
        print(f'Reason: {error["Cause"]}')
        exit()
    if format is None:
        print('\n')
        print(f'Getting Select ID {select_id}...')
        print('')
        print('*******************')
        print('Request details:')
        print(json.dumps(s.request.to_payload()))
        print('*******************')
        print('')
        geojson = json.loads(s.get_link_contents('geojson'))
        acqs = defaultdict(list)
        count = 0 
        for tile in geojson['features']:
            for feature in tile['properties']['best_matches']:
                count += 1
                acqs[feature['acquisition_id']].append(feature['date']) 
        print(f'{count} tiles identified in the following {len(acqs.keys())} acquisitions:')
        for acq, dates in acqs.items():
            print(f'{acq} ({dates[0]}) - {len(dates)} tiles')
        if len(acqs.keys()) > 20:
            print('WARNING: This Select has found more than 20 candidate acquisitions. Trying to order this with BBA will fail. Try more restrictive filters.')
        print('')
        print('Ordering this selection will use:')
        print(f'- {s.usage["fresh_imagery_sqkm"]} sqkm of fresh imagery')
        print(f'- {s.usage["standard_imagery_sqkm"]} sqkm of standard imagery')
        print(f'- {s.usage["training_imagery_sqkm"]} sqkm of training imagery')
        #print('')
        #print(f'As of {s.usage["usage_as_of"]}:')
        #print(f'{s.usage["available_sqkm"]} sqkm available')
        print('\n')
        exit()
    elif format in s.response.links.keys():
        print(s.get_link_contents(format))
    else:
        print(f'Unknown format {format}, try "html", "geojson", "geojsonl", or "stac"')

@click.command()
@click.argument('select_id', required=True)
@click.option('--format', help='Format of result file to get signed url for: `html`, `geojson`, geojsonl`, or `stac`')
def url(select_id, format):
    ''' Gets links for the Select result files for a Select of SELECT_ID. If no FORMAT is supplied, returns the base URL of the Select.
        If a FORMAT is supplied, returns a pre-signed URL to download the result of type FORMAT. '''
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    if not s.finished:
        print('Select has not finished running')
        sys.exit()
    if format is None:
        print(s.response.links['self'])
    elif format in s.response.links:
        print(s.get_signed_link(format)) 
    else:
        print(f'Unknown format {format}, try "html", "geojson", "geojsonl", or "stac"')
        

@click.command()
@click.argument('select_id', required=True)
def browse(select_id):
    ''' Downloads the HTML interactive viewer of the Select result SELECT ID to a temporary file and opens the
        map with the system web browser.
        When exited the temporary file is deleted. Use `max-ard select describe SELECT_ID --format html > my_map.html` 
        to save a local copy '''

    print('Fetching map...')
    try:
        s = Select.from_id(select_id)
    except Exception as e:
        print(str(e))
        sys.exit()
    if not s.finished:
        print('Select has not finished running')
        exit()
    prefix = f'ARD_Select_{select_id}-'
    with NamedTemporaryFile(prefix=prefix, suffix='.html', delete=False) as tmp:
        tmp.write(s.get_link_contents('html').encode(encoding='UTF-8'))
    # Windows won't let you open the file from browser while the context manager has it open
    # so we have to manually delete it
    webbrowser.open(f'file://{tmp.name}')
    time.sleep(5)
    os.unlink(tmp.name)

class NumericType(click.ParamType):
    name = "numeric"

    def convert(self, value, param, ctx):
        #strip commas if the got passed in a bbox
        value = value.replace(',', '')
        try:
            return float(value)
        except TypeError:
            self.fail(
                "expected string for int() conversion, got "
                f"{value!r} of type {type(value).__name__}",
                param,
                ctx,
            )
        except ValueError:
            self.fail(f"{value!r} is not a value that can be converted to a float", param, ctx)

NUM_TYPE = NumericType()

@click.command()
@click.option('--acq-id', 'acq_ids', multiple=True, help='Limit the Select to these acquisition IDs. Can be provided multiple times')
@click.option('--datetime', help='Limit the Select to a given date or date range')
@click.option('--intersects', help='Search for tiles that intersect this geometry in WKT format ')
@click.option('--bbox', nargs=4, type=NUM_TYPE, help='like `intersects`, but limits search to a WGS84 bounding box in format `--bbox XMIN YMIN XMAX YMAX`')
@click.option('--stack-depth', type=int, help='If provided, only return tiles where the stack depth can be fulfilled.')
@click.option('--filter', nargs=3, multiple=True, help='Add a filter statement in the form of `--filter <property> <operator> <value>. See docs for full filter syntax.')
@click.option('--image-age', 'image_age_category', multiple=True, type=click.Choice(['fresh', 'standard', 'training'], case_sensitive=False), help='Limit imagery to an image age category. Can be used more than once if age ranges are contiguous.')
@click.option('--min-cloud-free', type=NUM_TYPE, help='Shortcut to filter the minimum percentage of cloud-free cover. A value of 100 means an image must be 100% free of clouds.')
@click.option('--min-data', type=NUM_TYPE, help='Shortcut to filter the minimum percentage of valid pixels.')
def submit(**kwargs):
    ''' Submits a Select request to the server.

        There are numerous options and filter possibilities - consult the documentation for more information.

        At a minimum, you must supply at least one of the following:
        - a spatial filter - `intersects` or `bbox`
        - specific acquisition ID or IDs

        The Select service will try to compute results within 20 seconds. If it takes longer than 20 seconds the command 
        will return a job ID that you can check for completion with `max-ard select status SELECT_ID`
        '''
    min_cloud_free = kwargs.pop('min_cloud_free')
    min_data = kwargs.pop('min_data')
    filters = kwargs.pop('filter', [])

    query = defaultdict(dict)
    for prop, op, value in filters:
        # validate props and ops
        if op in ['eq', 'ne', 'gt', 'gte', 'lt', 'lte']:	
            try:	
                value = float(value)	
            except:	
                pass	
        # turn comma-delimited strings to lists	
        # and try to cast numbers if needed	
        elif op == 'in':	
            value = value.split(',')	
            value = [v.strip() for v in value]	
            try:	
                value = [float(v) for v in value]	
            except:	
                pass	
            	
        query[prop][op] = value 

    if min_cloud_free is not None:
        if min_cloud_free < 100:
            query['aoi:cloud_free_percentage']['gte'] = min_cloud_free
        else:
            query['aoi:cloud_free_percentage']['eq'] = 100

    if min_data is not None:
        if min_data < 100:
            query['aoi:data_percentage']['gte'] = min_data
        else:
            query['aoi:data_percentage']['eq'] = 100

    kwargs['query'] = query
    s = Select(**kwargs)

    print(f'Request payload: {s.request.to_payload()}')
    try:
        s.submit()
    except BadARDRequest as e:
        print(f'There was an error in your request: {e}')
        exit()
    if s.finished:
        print(f'Select {s.select_id} has completed')
    else:
        print(f'Select {s.select_id} is still running')
        print(f'Run `max-ard select status {s.select_id}` to check the status')
        print('When the status is "SUCCEEDED" you can: ')
    print(f'Run `max-ard select describe {s.select_id}` to see a basic overview of the results')
    print(f'Run `max-ard select browse {s.select_id}` to launch a browser-based map view of the results')
    print('Run `max-ard select` to see all the `select` commands')


        


select.add_command(status)
select.add_command(describe)
select.add_command(browse)
select.add_command(submit)
select.add_command(url)
